import PropTypes from 'prop-types';
import { createContext, useRef, useCallback, useMemo, useContext, useState, useEffect } from 'react';
import { jsx } from 'react/jsx-runtime';

let StatusEnum;

(function (StatusEnum) {
  StatusEnum["ERROR"] = "error";
  StatusEnum["IDLE"] = "idle";
  StatusEnum["LOADING"] = "loading";
  StatusEnum["SUCCESS"] = "success";
})(StatusEnum || (StatusEnum = {}));

const KEY_IS_NOT_STRING_ERROR = 'Key should be a string';
const DEFAULT_MAX_CONCURRENT_REQUESTS = 20;

class DataLoader {
  static maxConcurrent = DEFAULT_MAX_CONCURRENT_REQUESTS;
  static started = 0;
  static cachedData = {};
  static queue = {};
  isCalled = false;
  isCancelled = true;
  status = StatusEnum.IDLE;
  observers = [];
  loadCount = 0;
  isFirstLoading = true;

  constructor(args) {
    this.key = args.key;
    this.method = args.method;

    if (args.enabled) {
      this.status = StatusEnum.LOADING;
    }

    this.data = DataLoader.cachedData[this.key];
    if (args.notifyChanges) this.observers.push(args.notifyChanges);
  }

  notifyChanges() {
    if (this.timeout) {
      clearTimeout(this.timeout);
    }

    this.timeout = setTimeout(() => {
      this.observers.forEach(observer => observer());
    });
  }

  getData() {
    return DataLoader.cachedData[this.key];
  }

  tryLaunch = async () => {
    if (DataLoader.started < DataLoader.maxConcurrent) {
      DataLoader.started += 1;
      DataLoader.queue[this.key] = this.launch();
    } else {
      DataLoader.queue[this.key] = new Promise(resolve => {
        setTimeout(resolve);
      }).then(this.tryLaunch);
    }

    return DataLoader.queue[this.key];
  };
  load = async (force = false) => {
    if (force || !this.isCalled) {
      this.isCalled = true;

      if (this.status !== StatusEnum.LOADING) {
        this.status = StatusEnum.LOADING;
        this.notifyChanges();
      }

      DataLoader.queue[this.key] = this.tryLaunch();
    }

    return DataLoader.queue[this.key];
  };
  launch = async () => {
    try {
      this.isCancelled = false;
      this.loadCount += 1;
      const data = await this.method();

      if (!this.isCancelled) {
        DataLoader.cachedData[this.key] = data;
        this.status = StatusEnum.SUCCESS;
        this.data = data;
        this.error = undefined;
        this.dataUpdatedAt = Date.now();
      }

      this.isCalled = false;
      this.isFirstLoading = false;
      DataLoader.started -= 1;
      delete DataLoader.queue[this.key];
      this.notifyChanges();
      return data;
    } catch (error) {
      if (!this.isCancelled) {
        this.status = StatusEnum.ERROR;
        this.error = error;
      }

      this.isCalled = false;
      this.isFirstLoading = false;
      DataLoader.started -= 1;
      delete DataLoader.queue[this.key];
      this.notifyChanges();

      if (!this.isCancelled) {
        throw error;
      }

      return undefined;
    }
  };

  clearData() {
    DataLoader.cachedData[this.key] = undefined;
  }

  cancel() {
    DataLoader.started -= 1;
    delete DataLoader.queue[this.key];
    this.isCancelled = true;
    this.status = StatusEnum.IDLE;
  }

  addObserver(observer) {
    this.observers.push(observer);
  }

  removeObserver(observer) {
    const index = this.observers.indexOf(observer);
    if (index > -1) this.observers.splice(index, 1);
  }

}

const DataLoaderContext = createContext(undefined);

const DataLoaderProvider = ({
  children,
  cacheKeyPrefix,
  onError,
  maxConcurrentRequests
}) => {
  const requestsRef = useRef({});
  const computeKey = useCallback(key => `${cacheKeyPrefix ? `${cacheKeyPrefix}-` : ''}${key}`, [cacheKeyPrefix]);
  const getRequest = useCallback(key => requestsRef.current[computeKey(key)], [computeKey]);
  const addRequest = useCallback((key, args) => {
    if (DataLoader.maxConcurrent !== maxConcurrentRequests) {
      DataLoader.maxConcurrent = maxConcurrentRequests;
    }

    if (key && typeof key === 'string') {
      const newRequest = new DataLoader({ ...args,
        key: computeKey(key)
      });
      requestsRef.current[newRequest.key] = newRequest;
      return newRequest;
    }

    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey, maxConcurrentRequests]);
  const getOrAddRequest = useCallback((key, args) => {
    const requestFound = getRequest(key);

    if (!requestFound) {
      return addRequest(key, args);
    }

    return requestFound;
  }, [addRequest, getRequest]);
  const clearCachedData = useCallback(key => {
    if (typeof key === 'string') {
      if (requestsRef.current[computeKey(key)]) {
        requestsRef.current[computeKey(key)].clearData();
      }
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey]);
  const clearAllCachedData = useCallback(() => {
    Object.values(requestsRef.current).forEach(request => {
      request.clearData();
    });
  }, []);
  const reload = useCallback(async key => {
    if (key && typeof key === 'string') {
      await getRequest(key)?.load(true);
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [getRequest]);
  const reloadAll = useCallback(async () => {
    await Promise.all(Object.values(requestsRef.current).map(request => request.load(true)));
  }, []);
  const getCachedData = useCallback(key => {
    if (key) {
      return getRequest(key)?.getData();
    }

    return Object.values(requestsRef.current).reduce((acc, request) => ({ ...acc,
      [request.key]: request.getData()
    }), {});
  }, [getRequest]);
  const getReloads = useCallback(key => {
    if (key) {
      return getRequest(key) ? () => getRequest(key).load(true) : undefined;
    }

    return Object.entries(requestsRef.current).reduce((acc, [requestKey, {
      load
    }]) => ({ ...acc,
      [requestKey]: () => load(true)
    }), {});
  }, [getRequest]);
  const value = useMemo(() => ({
    addRequest,
    cacheKeyPrefix,
    clearAllCachedData,
    clearCachedData,
    getCachedData,
    getOrAddRequest,
    getReloads,
    getRequest,
    onError,
    reload,
    reloadAll
  }), [addRequest, cacheKeyPrefix, clearAllCachedData, clearCachedData, getCachedData, getOrAddRequest, getRequest, getReloads, onError, reload, reloadAll]);
  return jsx(DataLoaderContext.Provider, {
    value: value,
    children: children
  });
};

DataLoaderProvider.propTypes = {
  cacheKeyPrefix: PropTypes.string,
  children: PropTypes.node.isRequired,
  maxConcurrentRequests: PropTypes.number,
  onError: PropTypes.func
};
DataLoaderProvider.defaultProps = {
  cacheKeyPrefix: undefined,
  maxConcurrentRequests: DEFAULT_MAX_CONCURRENT_REQUESTS,
  onError: undefined
};
const useDataLoaderContext = () => useContext(DataLoaderContext);

const marshalQueryKey = queryKey => Array.isArray(queryKey) ? queryKey.filter(Boolean).map(subKey => {
  if (subKey instanceof Date) {
    return subKey.toISOString();
  }

  return subKey?.toString();
}).join('.') : queryKey?.toString();

function useDataLoader(key, method, {
  enabled = true,
  onError,
  onSuccess,
  keepPreviousData = false,
  needPolling = true,
  pollingInterval,
  initialData,
  dataLifetime
} = {}) {
  const {
    getOrAddRequest,
    onError: onGlobalError
  } = useDataLoaderContext();
  const methodRef = useRef(method);
  const onSuccessRef = useRef(onSuccess);
  const onErrorRef = useRef(onError ?? onGlobalError);
  const needPollingRef = useRef(needPolling);
  const [, setCounter] = useState(0);
  const forceRerender = useCallback(() => {
    setCounter(current => current + 1);
  }, []);
  const queryKey = useMemo(() => marshalQueryKey(key), [key]);
  const request = getOrAddRequest(queryKey, {
    enabled,
    method: methodRef.current
  });
  useEffect(() => {
    request.addObserver(forceRerender);
    return () => {
      request.removeObserver(forceRerender);
    };
  }, [request, forceRerender]);
  const needLoad = useMemo(() => !!(enabled && (!request.dataUpdatedAt || !dataLifetime || request.dataUpdatedAt && dataLifetime && request.dataUpdatedAt + dataLifetime < Date.now())), [enabled, request.dataUpdatedAt, dataLifetime]);
  const optimisticIsLoadingRef = useRef(needLoad);
  const previousDataRef = useRef(request.data);
  const isLoading = request.status === StatusEnum.LOADING || optimisticIsLoadingRef.current;
  const isSuccess = request.status === StatusEnum.SUCCESS;
  const isError = request.status === StatusEnum.ERROR;
  const isIdle = request.status === StatusEnum.IDLE && !enabled;
  const isPolling = !!(pollingInterval && (typeof needPolling === 'function' && (request.isFirstLoading || needPolling(request.data)) || typeof needPolling !== 'function' && needPolling));
  const reload = useCallback(() => request.load(true).then(onSuccessRef.current).catch(onErrorRef.current), [request]);
  useEffect(() => {
    needPollingRef.current = needPolling;
  }, [needPolling]);
  useEffect(() => {
    request.method = method;
  }, [method, request]);
  useEffect(() => {
    onSuccessRef.current = onSuccess;
  }, [onSuccess]);
  useEffect(() => {
    onErrorRef.current = onError ?? onGlobalError;
  }, [onError, onGlobalError]);
  useEffect(() => {
    if (keepPreviousData) {
      previousDataRef.current = request.data;
    }
  }, [request.data, keepPreviousData]);
  useEffect(() => {
    if (needLoad) {
      request.load().then(onSuccessRef.current).catch(onErrorRef.current);
    }

    optimisticIsLoadingRef.current = false;
  }, [needLoad, request]);
  useEffect(() => {
    let interval;

    if (pollingInterval) {
      interval = setInterval(() => {
        if (needPollingRef.current && typeof needPollingRef.current === 'function' && needPollingRef.current(request.data) || typeof needPollingRef.current !== 'function' && needPollingRef.current && !request.isCalled) {
          request.load(true).then(onSuccessRef.current).catch(onErrorRef.current);
        }
      }, pollingInterval);
    }

    return () => {
      if (interval) clearInterval(interval);
    };
  }, [pollingInterval, request]);
  return {
    data: !request.isFirstLoading ? request.data : initialData,
    error: request.error,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    previousData: previousDataRef.current,
    reload
  };
}

const usePaginatedDataLoader = (key, method, {
  enabled = true,
  initialData,
  keepPreviousData = true,
  onError,
  onSuccess,
  pollingInterval,
  dataLifetime,
  needPolling,
  initialPage,
  perPage = 1
} = {}) => {
  const [data, setData] = useState({});
  const [page, setPage] = useState(initialPage ?? 1);
  const keyPage = useMemo(() => [key, ['page', page]].flat(), [key, page]);
  const pageMethod = useCallback(() => method({
    page,
    perPage
  }), [method, page, perPage]);
  const {
    data: pageData,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    reload,
    error
  } = useDataLoader(keyPage, pageMethod, {
    dataLifetime,
    enabled,
    initialData,
    keepPreviousData,
    needPolling,
    onError,
    onSuccess,
    pollingInterval
  });
  const goToNextPage = useCallback(() => {
    setPage(current => current + 1);
  }, []);
  const goToPreviousPage = useCallback(() => {
    setPage(current => current > 1 ? current - 1 : 1);
  }, []);
  const goToPage = useCallback(newPage => {
    setPage(newPage > 1 ? newPage : 1);
  }, []);
  useEffect(() => {
    setData(current => {
      if (pageData !== current[page]) {
        return { ...current,
          [page]: pageData
        };
      }

      return current;
    });
  }, [pageData, page]);
  useEffect(() => {
    setPage(1);
    setData({});
  }, [key]);
  return {
    data,
    error,
    goToNextPage,
    goToPage,
    goToPreviousPage,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    page,
    pageData,
    reload
  };
};

export { DataLoaderProvider, useDataLoader, useDataLoaderContext, usePaginatedDataLoader };
